//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* curve.cpp: curves for fitting and more                                 *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <iostream>
#include <boost/regex.hpp>
#include <boost/format.hpp>
#include <stdlib.h>
#include <math.h>
#include <lmmin.h>

using boost::format;
using namespace std;

#include "mystd.h"
#include "olm.h" // includes also curve.h
#include "expr.h"
#include <readln.h>
#include <ask_simple_value.h>
#include "xax_lex.h"

//! local functions (wouldn't suffice "static" ??)

namespace NCurveFile
{
    void IntegrateFile(
        const COlc *fin, const int k, const string& mode,
        const int ip, const string& name, const CCoord& ico );
    double NumericIntegral(
        const COlc *fc, const CCurve *c, const int k, const int j,
        const double low, const double hig );
    // parameters for numeric integration:
    double numint_bound_low=-INFINITY, numint_bound_hig=INFINITY;
    // tuning parameters:
    int numint_mode=4;
    int numint_cvmode=-1;
    double numint_epsabs=1e-10, numint_epsrel=1e-10;
};

bool allow_slow_conv = true;

//***************************************************************************//
//* generic operations on curves                                            *//
//***************************************************************************//


//! Initialize curve. Used when creating or loading a curve file.

void CCurve::Clear(void)
{
    P.clear();
    z.clear();
    Quality[0] = -1;
    Quality[1] = -1;
    Quality[2] = -1;
}


//***************************************************************************//
//* Generic operations on curve files                                       *//
//***************************************************************************//


//! Create curve files.

void NCurveFile::Create( const string& cmd )
{
    static int default_no=0;

    // dialog:
    string expr = sask( "Fit function ?" );
    if( expr=="" ) return;
    COlc f;
    setFunction( &f, expr );
    vector<double> P( f.nPar(), 1.0 );

    // use f to create output file(s)
    if        ( cmd=="cc" ){
        // adapt one output curve file for each selected data file
        NOlm::IterateD fiter;
        NOlm::SelNew();
        const COld *fd;
        while( (fd=fiter()) ) {
            COlc fout = f;
            fout.name = string("fit_") + fd->name;
            fout.kd = fiter.SelNo();
            fout.xco = CCoord( "ip", "" );
            fout.yco = CCoord( "p[ip]", "" );
            fout.ZCo = fd->ZCo;
            fout.RPar = fd->RPar;
            fout.lDoc.push_back( "cc " + expr + " # z from " + fd->name );
            for (uint j=0; j<fd->nScan(); j++) {
                CCurve sout;
                sout.z = fd->VS[j].z;
                sout.P = P;
                fout.VC.push_back( sout );
            }
            NOlm::OloAdd( &fout );
        }
    } else if ( cmd=="cca" ) {
        // create just one "artificial" curve file with no data reference
        COlc fout( f );
        fout.name = string("curve") + strg(default_no++);
        fout.xco = CCoord( "x", "" );
        fout.yco = CCoord( "y", "" );
        fout.kd = -1;
        fout.lDoc.push_back( "cca " + expr );
        CCurve sout;
        sout.P = P;
        fout.VC.push_back( sout );
        NOlm::SelNew();
        NOlm::OloAdd( &fout );
    } else if ( cmd=="cm" ){
        // modify function for existing file
        NOlm::IterateC fiter;
        COlc *fc;
        while( (fc=fiter()) ) {
            uint nParOld = fc->nPar();
            fc->lDoc.push_back( "ccm " + expr );
            fc->expr = f.expr;
            fc->T = f.T;
            for( uint ip=nParOld; ip<f.T->npar(); ip++ ){
                fc->PCo.push_back( CCoord("p" + strg(ip), "") );
                fc->Fixed.push_back( CList() );
            }
            for (uint j=0; j<fc->nScan(); j++)
                for( uint ip=nParOld; ip<f.T->npar(); ip++ )
                    fc->VC[j].P.push_back( 1.0 );
        }
    } else
        throw "unknown command " + cmd;
}


//! Parse expression, and set auxiliary parameters.
//  Used here (Create) and by when reading from disk (NFileIn::Load).

void NCurveFile::setFunction( void* _f, const string& expr )
// for obscure reasons, it is not possible to define arg1 as COlc*;
// g++ then thinks COlc is part of the present namespace.
{
    COlc* f; f = (COlc*) _f;
    // parse expression
    PTree T;
    user_xaxparse( expr.c_str(), &T );
    if( !T->has_farg )
        cout << "WARNING: curve has no formal argument t\n";
    // set various records of *f
    f->expr = expr;
    f->T = T;
    for( uint ip=0; ip<T->npar(); ip++ ){
        f->PCo.push_back( CCoord("p" + strg(ip), "") );
        f->Fixed.push_back( CList() );
    }
    f->weight_y_log = false;
    f->cv_intmod = NCurveFile::numint_cvmode;
    f->cv_epsabs = NCurveFile::numint_epsabs;
    f->cv_epsrel = NCurveFile::numint_epsrel;
}


//! Display parameter list for set of online curves.

void NCurveFile::ShowPar(void)
{
    NOlm::IterateC fiter;
    const COlc *fin;
    while( (fin=fiter()) )
        cout << fin->pInfoCat();
}

//! Return parameter list for one online curve.

vector<string> COlc::pInfo() const
{
#define PINFO_WRD_SIZ 15
    char wrd[PINFO_WRD_SIZ];
    string lin;
    vector<string> ret;
    // file-wide info:
    ret.push_back( expr );
    if( kd!=-1 )
        ret.push_back( "refers to data:  " + strg(kd) );
    if( kconv!=-1 )
        ret.push_back( "convolute with: " + strg(kconv) );
    ret.push_back( "weight(ylog):    " + strg(weight_y_log) );
    // z's and parameters:
    lin = "j   ";
    for ( uint iz=0; iz<nZ(); iz++ ) {
        snprintf( wrd, PINFO_WRD_SIZ, "z%i           ", iz );
        lin += wrd;
    }
    for ( uint ip=0; ip<nPar(); ip++ ) {
        snprintf( wrd, PINFO_WRD_SIZ, "p%d            ", ip );
        lin += wrd;
    }
    ret.push_back( lin );
    // z's and parameters per scan:
    for ( uint j=0; j<nScan(); j++ ) {
        lin = "";
        snprintf( wrd, PINFO_WRD_SIZ, "%-2i%c ", j,
                  ProtectedScans.contains(j) ? '/' : ' ' );
        lin += wrd;
        for ( uint iz=0; iz<nZ(); iz++ ) {
            snprintf( wrd, PINFO_WRD_SIZ, "%-12.7g ", VC[j].z[iz] );
            lin += wrd;
        }
        for ( uint ip=0; ip<nPar(); ip++ ) {
            snprintf( wrd, PINFO_WRD_SIZ, "%c%-12.7g ",
                      Fixed[ip].contains(j) ? 'x':' ', VC[j].P[ip] );
            lin += wrd;
        }
        ret.push_back( lin );
    }
    return ret;
}

string COlc::pInfoCat() const
{
    vector<string> vs = pInfo();
    string ret = "";
    for( int i=0; i<vs.size(); ++i )
        ret += vs[i] + "\n";
    return ret;
}

//! Overwrite curve parameters.

void NCurveFile::OprPar(string llabel) 
{
    CRef lref(llabel);
    if( !lref.defined() )
        throw "invalid left-hand side " + llabel;
    else if (lref.var!=CRef::_FP)
        throw string( "PROGRAM ERROR: OprPar: im falschen Film");

    string expr = sask("Function ?");
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateC fiter;
    COlc *f;
    double val;
    while( (f=fiter()) ) {
        uint k = fiter.SelNo();
        if ( f->ProtectedScans.size() )
            cout << "some scans are protected\n";
        if( lref.num>=f->nPar() )
            throw string( "invalid left-hand side: not that many parameters" );
        f->lDoc.push_back( "o" + llabel + " " + expr);
        for (uint j=0; j<f->nScan(); j++) {
            if ( f->ProtectedScans.contains( j ) )
                continue;
            T->tree_point_val( &val, k, j );
            f->VC[j].P[lref.num] = val;
        }
    }
}


//! Fix or unfix curve parameters.

void NCurveFile::SetFixed()
{
    // Everything is driven by one input string
    string lin;
    while ( true ) {
        lin = sask( "Change fix status", "help" );
        if        ( lin=="h" || lin=="help" ) {
            cout <<
                "cx input:\n"
                "  blank separated list of instructions\n"
                "instruction:\n"
                "  <ip>+         # to make parameter number ip fixed\n"
                "  <ip>-         # to release the fixation\n"
                "  <ip>=<j-list> # to fix the parameter for scans in j-list\n";
            continue;
        } else if ( lin=="q" )
            return;
        break;
    }

    boost::regex re1( "\\s+" );
    boost::regex re2( "(\\d+)(\\+|\\-|\\=)(.*)" );

    // Split input string in words
    boost::sregex_token_iterator _wrd( lin.begin(), lin.end(), re1, -1 );
    while ( _wrd!=boost::sregex_token_iterator() ) {
        // Parse one instruction
        string wrd = *_wrd++;
        boost::smatch tokens;
        if( !boost::regex_match( wrd, tokens, re2 ) )
            throw "invalid instruction "+wrd;
        if ( tokens[2]!="=" && tokens[3]!="" )
            throw "invalid instruction "+wrd;
        int ip;
        if( !mystd::any2int( tokens[1], &ip ) )
            throw string( "BUG: unreadable ip in spite of match" );
        // Loop over selected curve files
        NOlm::IterateC fiter;
        COlc *f;
        while( (f=fiter()) ) {
            if( ip>=f->nPar() )
                throw string( "invalid parameter number" );
            if      ( tokens[2] == "+" )
                f->Fixed[ip] = CList( 0, f->nScan()-1 );
            else if ( tokens[2] == "-" )
                f->Fixed[ip] = CList( );
            else if ( tokens[2] == "=" )
                f->Fixed[ip] = CList( tokens[3], 0, f->nScan()-1 );
        }
    }
}


//! Set data file or convolution file.

void NCurveFile::SetFileReference( const string& which )
// which = "d"  set data file
//         "v"  set convolution file
//         "v-" switch convolution off
{
    string expr;
    PTree T;
    if( which=="v-" )
        ; // no preparation
    else {
        if     ( which=="d" )
            expr = sask( "Fit data file" );
        else if( which=="v" )
            expr = sask( "Convolute with file" );
        else
            throw "BUG: unexpected which="+which;
        if (expr!="")
            user_xaxparse( expr.c_str(), &T );
    }
    COlc *fc;
    NOlm::IterateC fiter;
    while ( (fc=fiter()) ) {
        uint k = fiter.SelNo();
        if     ( which=="v-" )
            fc->kconv = -1;
        else if( expr=="" ) {
            cout << format( "curve %-2d -> data %-2d" ) % k % fc->kd;
            if( fc->kconv!=-1 )
                cout << format( "; convolute with %-2d" ) % k % fc->kconv;
            cout << "\n";
        } else {
            uint lval;
            CContext ctx( k );
            T->tree_uival( &lval, &ctx );
            if( lval>=NOlm::MOM.size() )
                throw "no data file " + strg(lval);
            if( ! NOlm::MOM[lval].D() )
                throw "file " + strg(lval) + " isn't a data file";
            if     ( which=="d" )
                fc->kd = lval;
            else if( which=="v" )
                fc->kconv = lval;
        }
    }
}


//! Set auxiliary parameters for numeric convolution.

void NCurveFile::SetConvTuningPars( string which )
{
    // per-file ???
    throw string( "convolution tuning parameters not operational" );
    
    static int mode = -1;
    static double epsabs = 1e-10, epsrel = 1e-10;
    COlc *fc;
    NOlm::IterateC fiter;
    if        ( which=="?" ) {
        while ( (fc=fiter()) ) {
            printf( "file %i:\n", fiter.SelNo() );
            printf( "   cv  convolute with:   %i\n", fc->kconv );
            printf( "   cvm convolution mode: %i\n", fc->cv_intmod );
            printf( "   cva eps_abs:          %g\n", fc->cv_epsabs );
            printf( "   cvr eps_rel:          %g\n", fc->cv_epsrel );
        }
    } else if ( which=="m" ) {
        mode = iask( "Convolution mode (-1: Riemann, 0..7: smooth..singular" );
        if( mode<-1 || mode>7 ){
            mode = -1;
            return;
        }
        while ( (fc=fiter()) )
            fc->cv_intmod = mode;
    } else if ( which=="a" ) {
        epsabs = dask( "Eps abs for convolution", epsabs );
        while ( (fc=fiter()) )
            fc->cv_epsabs = epsabs;
    } else if ( which=="r" ) {
        epsrel = dask( "Eps rel for convolution", epsrel );
        while ( (fc=fiter()) )
            fc->cv_epsrel = epsrel;
    }
}

//! Set auxiliary parameters (weighing, ..).

void NCurveFile::SetProperties(string which)
{
    COlc *fc;
    NOlm::IterateC fiter;
    while((fc=fiter())) {
        if (which=="yli")
            fc->weight_y_log = false;
        else if (which=="ylg")
            fc->weight_y_log = true;
        else {
            printf("! unknown property");
            return;
        }
    }
}


//***************************************************************************//
//* Least-squares fitting                                                   *//
//***************************************************************************//


//! Data transfer between frida and lm_minimize callback functions.

typedef struct {
    const COld *fd;
    COlc *fc;
    uint k;
    uint j;
    double timeout;
} FitDatTyp;


//! Callback routine for use in lm_minimize: function evaluation.

void globalEvaluate( double* par, int m_dat, double* fvec, 
                     void *data, int *info )
// arguments (required by lm_minimize):
//    par   r/w  function parameters, to be optimized
//    m_dat r    number of data points
//    fvec  w    vector of residues, to be minimized
//    data  r    pointer to arbitrary data, here defined as FitDatTyp
//    info  w    can be use to terminate the minimization
{
    try {
        FitDatTyp *mydata;
        mydata = (FitDatTyp*) data;
        if( mydata->timeout && time(NULL) > mydata->timeout )
            throw string( "reached fit time out" );
        COlc *fc = mydata->fc;
        const COld *fd = mydata->fd;
        CCurve *c = &(fc->VC[ mydata->j ]);
        
        uint ip = 0;
        for( uint i=0; i<fc->nPar(); ++i )
            if ( !(fc->Fixed[i].contains( mydata->j )) )
                c->P[i] = par[ip++];
        
        const CScan *s = &(fd->VS[mydata->j]);
        vector<double> fit;
        fc->T->tree_curve_val_vec( &fit, s->x, mydata->k, mydata->j );
        
        for ( int i=0; i<m_dat; i++) {
            if( fit[i]==INFINITY ) // TEMPORARY UNTIL exceptions can be
                                   // thrown everywhere
                throw string( "function evaluation return INFINITY" );

            if (!fc->weight_y_log) {
                fvec[i] = s->y[i] - fit[i];
/*
                fvec[i] *= fvec[i]; // TEMPORARY
                fvec[i] *= fvec[i]; // TEMPORARY
                fvec[i] *= fvec[i]; // TEMPORARY
*/
            } else {
                if ( s->y[i] < 1e-99 || fit[i] < 1e-99 ) {
                    fvec[i] = 0; // just don't fit
                } else {
                    fvec[i] = log( s->y[i] / fit[i] );
                }
            }
        }
    } catch ( string& s ) {
        cerr << "\n" << s << "\n";
        *info = -1;
    }
}


//! Callback routine for use in lm_minimize: monitor fit progress.

void globalPrintout( int n_par, double* par, int m_dat, double* fvec, 
                     void *data, int iflag, int iter, int nfev )
// arguments (required by lm_minimize; see also globalEvaluate):
//       iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
//       iter  : outer loop counter
//       nfev  : number of calls to *evaluate
{
}


//! NFitTune holds auxiliary parameters for the fit routine

namespace NFitTune {
    double Epsilon = 1.e-14;
    double Factor=100.;
    int nCall=200;
    int nPrint=0;
    int mFitMetric=1;
    double maxtime = 3;
};


//! Set auxiliary parameters. Directly called from main program.

void NCurveFile::SetFitTuningPars( string which )
{
    if      ( which=="?" )
        printf( "Fit setup:\n   nCall=%d\n   nPrint=%d\n   Factor=%g\n"
                "   Eps=%g\n   Metric=%i\n",
                NFitTune::nCall,
                NFitTune::nPrint,
                NFitTune::Factor,
                NFitTune::Epsilon,
                NFitTune::mFitMetric );
    else if ( which=="p" )
        NFitTune::nPrint = iask( "Print frequency", NFitTune::nPrint );
    else if ( which=="c" )
        NFitTune::nCall = iask( "Max no of calls", NFitTune::nCall );
    else if ( which=="f" )
        NFitTune::Factor = dask( "NFitTune::Factor", NFitTune::Factor );
    else if ( which=="e" )
        NFitTune::Epsilon = dask( "NFitTune::Epsilon", NFitTune::Epsilon );
    else if ( which=="m" ){
        // metric (0) is the integer indicator returned by lm_minimize;
        // it cannot be chosen here
        printf( "Fit metric:\n"
                "(1) coefficient of determination\n"
                "(2) mean deviation\n" );
        int ret = iask( "Option", NFitTune::mFitMetric );
        if( ret<1 || ret> N_FIT_METRICS )
            throw string( "invalid option" );
        NFitTune::mFitMetric = ret;
    } else if ( which=="t" )
        NFitTune::maxtime = dask( "Max duration of fits (in s)",
                                  NFitTune::maxtime );
}


//! Fit.

void NCurveFile::Fit( bool _allow_slow_conv )
{
    COlc *fc;
    COld *fd;
    CScan *D;
    CCurve *C;
    vector<double> Par;
    lm_control_type control;
    FitDatTyp data;
    double ysum1, ysum2, ydev2;
    allow_slow_conv = _allow_slow_conv;

    NOlm::IterateC fiter;
    while((fc=fiter())) {
        data.k = fiter.SelNo();
        if( !fc->T->has_farg )
            throw string( "curve has no formal argument t" );
        if( fc->kd<0 || fc->kd>=NOlm::MOM.size() )
            throw string( "number of data file out of range" );
        if (!(fd=NOlm::MOM[fc->kd].D()))
            throw string( "is not a data file" );
        if (fc->nScan() != fd->nScan())
            throw string( "data file and curve file have "
                          "different number of scans" );
        if ( fc->ProtectedScans.size() )
            cout << "some scans are protected\n";
        fc->lDoc.push_back( "fitted to file " + strg(fc->kd) + " " + fd->name );
        
        data.fc = fc;
        data.fd = fd;

        uint np = fc->T->npar();
                
        lm_initialize_control( &control );
        control.epsilon = NFitTune::Epsilon;
        control.stepbound = NFitTune::Factor;
        control.maxcall = NFitTune::nCall;

        for (uint j=0; j<fc->nScan(); j++) {
            if ( fc->ProtectedScans.contains( j ) )
                continue;
            data.j = j;
            if( fiter.size()>1 )
                printf("%3d", data.k);
            printf("%3d", j);
            D = &(fd->VS[j]);
            C = &(fc->VC[j]);

            // determine number of free parameters
            uint npfree = 0;
            for ( uint ip=0; ip<np; ++ip )
                if( !fc->Fixed[ip].contains(j) )
                    ++npfree;
            if (npfree<=0)
                continue;
            Par.resize(npfree);
            uint ipf=0;
            for ( uint ip=0; ip<np; ip++ ) {
                if ( !fc->Fixed[ip].contains(j) )
                    Par[ipf++] = C->P[ip];
            }
                                        
            if ( NFitTune::nPrint ) printf("\n");

            data.timeout = time(NULL) + NFitTune::maxtime;
            lm_minimize( D->size(), npfree, &(Par[0]), 
                         globalEvaluate, globalPrintout, &data, &control );
            if( control.info==10 )
                throw string( "exception in fit function" );

            for ( uint i=0; i<npfree; i++ )
                printf(" %12g", Par[i]);

            C->Quality[0] = (double) control.info;

            // calculate coefficient of determination R^2:
            uint nd = fd->nPts(j);
            ysum1 = 0, ysum2 = 0;
            for( uint i=0; i<nd; ++i ){
                ysum1 += D->y[i];
                ysum2 += SQR( D->y[i] );
            }
            ydev2 = ysum2 - SQR( ysum1 )/nd; // = nd * <(y-<y>)^2>
            C->Quality[1] = ydev2!=0 ? 1 - SQR(control.fnorm) / ydev2 : -1;
                        
            // mean deviation:
            C->Quality[2] = SQR(control.fnorm) / nd;

            printf( " (%3d->%9.7g)",
                    control.nfev, C->Quality[NFitTune::mFitMetric] );

            // print status:
            printf( "%s\n", lm_shortmsg[ control.info ] );
        }
    }
    allow_slow_conv = true;
}


//***************************************************************************//
//* Generate integral file (extract parameters, integrate, ...)             *//
//***************************************************************************//

//! Data transfer between frida and gsl_function.

typedef struct {
    const COlc *fc;
    uint k;
    uint j;
} EvalDatTyp;


//! Generic callback routine for tree evaluation by gsl_function.

double myeval( double x, void* data )
{
    EvalDatTyp *mydata = (EvalDatTyp*) data;
    return mydata->fc->T->tree_curve_val_sca( x, mydata->k, mydata->j );
}


//! display or change tuning parameters for numeric integration

void NCurveFile::SetIntTuningPars( string which )
{
    if      ( which=="?" )
        printf( "Numeric integration setup:\n"
                "   cna epsabs=%g\n"
                "   cnr epsrel=%g\n"
                "   cnm external integration mode: %d\n"
                "   cnv convolution integral mode: %d\n",
                numint_epsabs, numint_epsrel, numint_mode, numint_cvmode );
    else if ( which=="a" )
        numint_epsabs = dask( "Espilon abs", numint_epsabs );
    else if ( which=="r" )
        numint_epsrel = dask( "Epsilon rel", numint_epsrel );
    else if ( which=="m" ){
        cout << "Numeric integration modes (ignored for infinite intervals):\n"
            "   0: non-adaptive Gauss-Kronod-87 for very smooth functions\n"
            "   1: adaptive Gauss-Kronod-61 for smooth functions\n"
            "   ...\n"
            "   6: adaptive Gauss-Kronod-15 for not so smooth functions\n"
            "   7: adaptive for functions with singularities\n";
        numint_mode = iask( "Option", numint_mode );
    } else if ( which=="v" ){
        cout << "Convolution integration modes:\n"
            "  -1: Riemann, steps given by resolution data\n"
            "   0: non-adaptive Gauss-Kronod-87 for very smooth functions\n"
            "   1: adaptive Gauss-Kronod-61 for smooth functions\n"
            "   ...\n"
            "   6: adaptive Gauss-Kronod-15 for not so smooth functions\n"
            "   7: adaptive for functions with singularities\n";
        numint_cvmode = iask( "Option", numint_cvmode );
    }
}

//! Outer wrapper: dialog, loop over files and parameters

void NCurveFile::IntegralProperty(void)
{
    if( NRead::stack_empty() )
        cout << "Generate integral file:\n"
            "   <list of numbers>   extract curve parameters\n"
            "   *                   extract all curve parameters\n"
            "   q0 | q1 | q2        extract fit metrics\n"
            "   i                   numeric integral\n";
    static string mode = "*";
    mode = wask( "Option", mode );
    if      ( mode==" " )
        return;
    else if ( mode=="i" ){
        if( NRead::stack_empty() )
            cout << "integration bounds: -inf,inf are allowed\n";
        numint_bound_low = dask( "Integrate from", numint_bound_low );
        numint_bound_hig = dask( "Integrate to", numint_bound_hig );
    }

    NOlm::IterateC fiter;
    bool reset_if_savable = true;
    const COlc* fin;
    while( (fin=fiter()) ){

        if( fin->ZCo.size()>0 && reset_if_savable ){
            NOlm::SelNew();
            reset_if_savable = false;
        }

        if( isdigit(mode[0]) || mode=="*" ){ // extract parameter(s)
            // decode list of parameter numbers
            CList parsel( mode, 0, fin->nPar()-1 );
            for( uint iv=0; iv<parsel.size(); ++iv ){
                int ip = parsel.V[iv];
                IntegrateFile( fin, fiter.SelNo(),
                               "p", ip, "p"+strg(ip), fin->PCo[ip] );
            }
        } else // other integrals
            IntegrateFile( fin, fiter.SelNo(), mode, -1, mode,
                           CCoord( "mode", "" ) );
    }
}


//! Extract exactly one integral from each curve of fin.

void NCurveFile::IntegrateFile( const COlc *fin, const int k,
                                const string& mode, const int ip,
                                const string& name, const CCoord& ico )
{
    COld fout;
    CScan S;

    uint nz = fin->ZCo.size();
    bool savable = ( nz > 0 );

    if( savable ){
        fout = COld( fin );
        if( fout.name.substr(0,4)=="fit_" )
            fout.name.replace( 0, 4, "" );
        fout.name.insert( 0, name + "_" );
        fout.xco = fin->ZCo.back();
        fout.yco = ico;
        if( fout.ZCo.size()<1 )
            throw string( "input has no z coord" );
        fout.ZCo.pop_back();
        fout.lDoc.push_back( "ci " + ( mode=="p" ? strg(ip) : mode ) );
        S.Clear();
        S.z = fin->VC[0].z;
        S.z.pop_back();
    }

    for (uint j=0; j<fin->nScan(); j++) {

        // determine integral value
        double val;
        if      ( mode=="p" )
            val = fin->VC[j].P[ip];
        else if ( mode=="q0" )
            val = fin->VC[j].Quality[0];
        else if ( mode=="q1" )
            val = fin->VC[j].Quality[1];
        else if ( mode=="q2" )
            val = fin->VC[j].Quality[2];
        else if ( mode=="i" )
            val = NumericIntegral( fin, &(fin->VC[j]), k, j,
                                   numint_bound_low, numint_bound_hig );
        else
            throw string("invalid mode");

        // print out ?
        if (!savable) {
            cout << fin->name << ": " << name << " = " << val << "\n";
            continue;
        }
        
        // store in integral scan
        S.push_xy( fin->VC[j].z[nz-1], val );
        // move z values
        if (nz>=2 &&
            (j+1)<fin->nScan() && 
            fin->VC[j+1].z[nz-2]!=fin->VC[j].z[nz-2]) {
            fout.VS.push_back(S);
            S.Clear();
            S.z = fin->VC[j+1].z;
            S.z.pop_back();
        }
    }
    if (savable) {
        fout.VS.push_back(S);
        NOlm::OloAdd(&fout);
    }
};


//! Numeric integration of one curve.

double NCurveFile::NumericIntegral(
    const COlc *fc, const CCurve *c, const int k, const int j,
    const double low, const double hig )
{
    EvalDatTyp data;
    data.fc = fc;
    // data.fd = NOlm::MOM[fc->kd].D();
    data.k = k;
    data.j = j;

    return mystd::Integrate( &myeval, &data, low, hig,
                             numint_mode, numint_epsabs, numint_epsrel );
}

//***************************************************************************//
//* Calculate cumulative distribution (needed for SPHERES PST simulation)   *//
//***************************************************************************//


//! Data transfer between frida and gsl_function.

typedef struct {
    const COlc *fc;
    uint k;
    uint j;
    double low;
    double hig;
} CumDisDatTyp;


namespace NCurveFile
{
    void DivideInterval( 
        vector<double> &y, const int ilow, const int ihig,
        CumDisDatTyp &data );
}

//! Calculate cumulative probability distribution on equidistant grid

/* Input:  `dis(x)`, a probability distribution.
   Output: `cum(r)`, the inverse of the cumulative probability distribution
     based on `dis`. It is implicitely understood that the abscissa of `cum`
     is an equidistant grid 2^m+1 points from 0 to 1.

   The resulting `cum` is used as follows. To draw random values x
   that are distributed according to `dis`:
   - Draw a random number r from a uniform distribution in the interval [0,1].
   - Calculate ilow=floor(r*2^m) and ihig = ilow+1.
   - Obtain x = cum(ilow) + (cum(ihig)-cum(ilow))*(r-ilow/2^m).
*/

void NCurveFile::CumulativeDistribution(void)
{
    static int nb = 3;
    nb = iask( "How many twofold divisions", nb );
    if( nb<1 ) return;

    static double low=-3, hig=+3;
    low = dask( "Lower bound for Phi^-1(r)", low );
    hig = dask( "Upper bound for Phi^-1(r)", hig );
    if( low>=hig ) return;

    NOlm::IterateC fiter;
    NOlm::SelNew();
    const COlc* fin;
    while( (fin=fiter()) ){
        COld fPhi;
        fPhi.name = "Phi_" + fin->name;
        fPhi.xco = CCoord( "r", "" );
        fPhi.yco = CCoord( "Phi^-1(r)", "" );
        fPhi.ZCo = fin->ZCo;
        fPhi.lDoc = fin->lDoc;
        fPhi.lDoc.push_back( "cumulative distribution" );
        
        for( uint j=0; j<fin->VC.size(); j++ ){
            const CCurve* C = &(fin->VC[j]);
            CScan S;
            S.z = C->z;
            S.x.resize( (1<<nb)+1 );
            for( int i=0; i<(1<<nb)+1; ++i )
                S.x[i] = ((double)i)/(1<<nb);
            S.y.resize( (1<<nb)+1 );
            S.y[0] = low;
            S.y[1<<nb] = hig;

            // recursive solution (acts on S.y):
            CumDisDatTyp data;
            data.fc = fin;
            data.k = fiter.SelNo();
            data.j = j;
            DivideInterval( S.y, 0, 1<<nb, data );

            fPhi.VS.push_back(S);
        }

        NOlm::OloAdd(&fPhi);
    }
}


//! Calculate Phi^-1(r) by recursively dividing Phi-intervals.

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>

double IntervalDifference( double mid, void *data )
{
    const CumDisDatTyp *mydata = (const CumDisDatTyp*) data;
    int j = mydata->j, k = mydata->k;
    const COlc *fc = mydata->fc;
    return NCurveFile::NumericIntegral( fc, &(fc->VC[j]), k, j,
                                        mydata->low, mid )
        -  NCurveFile::NumericIntegral( fc, &(fc->VC[j]), k, j,
                                        mid, mydata->hig );
}

void NCurveFile::DivideInterval(
    vector<double> &Y, const int ilow, const int ihig,
    CumDisDatTyp &data )
{
    if( ihig-ilow<=1 )
        return; // end of recursion

    // starting values for intersection:
    double low = Y[ilow], hig = Y[ihig], mid = (hig+low)/2;
    
    // integration limits for IntervalDifferences:
    data.low = low; data.hig = hig;

    // overhead for GSL root finder:
    const gsl_root_fsolver_type *root_solver = gsl_root_fsolver_brent;
    gsl_root_fsolver *s = gsl_root_fsolver_alloc( root_solver );
    gsl_function F;
    F.function = &IntervalDifference;
    F.params = &data;
    gsl_root_fsolver_set( s, &F, low, hig );
    int status, iter = 0, max_iter = 20;

    // iterative root finding:
    do {
        //printf( "%4d %4d %5d [%.7f, %.7f] %.7f\n",
        //        ilow, ihig, iter, low, hig, mid );
        iter++;
        status = gsl_root_fsolver_iterate (s);
        mid = gsl_root_fsolver_root (s);
        low = gsl_root_fsolver_x_lower (s);
        hig = gsl_root_fsolver_x_upper (s);
        status = gsl_root_test_interval( low, hig, 0, 1e-10 );
        // if (status == GSL_SUCCESS) printf ("Converged.\n");
    } while( status == GSL_CONTINUE && iter < max_iter );
    gsl_root_fsolver_free (s);
   
    // save result:
    int imid = (ilow+ihig)/2;
    Y[imid] = mid;

    // continue recursion:
    DivideInterval( Y, ilow, imid, data );
    DivideInterval( Y, imid, ihig, data );
}
