//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* expr.cpp: evaluate generic expressions                                 *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <math.h>
#include <iostream>
#include <stdlib.h> // for boost/shared_ptr

#include "mystd.h"
#include "olm.h" // for cross-references
#include "expr.h"
#include "readln.h"

// for calculator:
#include <ask_simple_value.h>
#include "readln.h"

bool debug = false;
extern bool allow_slow_conv;
#define DEB(args) if ( debug ) printf args


//***************************************************************************//
//* xax: a flex/bison based parser                                          *//
//***************************************************************************//

#include "xax_yacc.h"
#include "xax_lex.h"

int xaxerror( const char *s )  /* Called by yyparse on error */
{
    throw( string( "xaxerror: " ) + s );
    return 1;
}


//***************************************************************************//
//* class CTOut                                                             *//
//* (return type for tree/reference evaluation)                             *//
//***************************************************************************//

string CTOut::info() const
{
    string ret;
    if      ( typ==_OI )
        ret = "out(i:" + strg(d) + ")";
    else if ( typ==_OD )
        ret = "out(d:" + strg(d) + ")";
    else if ( typ==_OF )
        ret = "out(f:k=" + strg(cr.k) + ",j=" + strg(cr.j) + ")";
    else if ( typ==_OVD ) {
        ret = "out(v:n=" + strg(vd.size()) + ") [ ";
        for( uint i=0; i<vd.size(); ++i )
            ret += strg(vd[i]) + " ";
        ret += "]";
    }
    return ret;
}

//***************************************************************************//
//* class CContext                                                          *//
//***************************************************************************//

string CContext::info() const
{
    string ret = "dimension: ";
    switch( reqDim ) {
    case _1: ret += "scalar"; break;
    case _VI: ret += "vector(i)"; break;
    case _VT: ret += "vector(t)"; break;
    default:  ret += "undefined (BUG!)";
    }
/*
    ret += ", type: ";
    switch( TReqTyp ) {
    case _QI: ret += "int"; break;
    case _QD: ret += "dbl"; break;
    default:  ret += "undefined (BUG!)";
    }
*/
    ret += ", k: " + strg(k);
    ret += ", j: " + strg(j);
    ret += ", i: " + strg(i);
    ret += ", nv: " + strg(nv);
    return ret;
}

//***************************************************************************//
//* class CRef                                                              *//
//* (references, typically used as endpoints of CTree's)                    *//
//***************************************************************************//


CRef::CRef( string s, 
	    PTree _tk, PTree _tj, PTree _ti )
    // NOTE: As long as this constructor is called by the parser and not
    // by the lexical analyzer, string s may contain more than just the
    // parameter name. Everything after the parameter name must be ignored.
{
    num = 0;
    int isub;
    char c;
    int nsub = sscanf(s.c_str(), "%c%d", &c, &isub);
    if (s.substr(0,2)=="fm" && sscanf(s.substr(2).c_str(), "%d", &isub)==1 ) {
        var = _FM; // fit metric
        num = isub;
    } else if (nsub==2 && c=='p') {
        var = _FP; // fit parameter
        num = isub;
    } else if (nsub==2 && c=='z') {
        var = _Z;  // z value
        num = isub;
    } else if (s.substr(0,2)=="nj") {
        var = _NJ; // number of scans in file
    } else if (s.substr(0,2)=="ni") {
        var = _NI; // number of points in scan
    } else if (nsub==1 && c=='x') {
        var = _X;  // x value
    } else if (nsub==1 && c=='y') {
        var = _Y;  // y value
    } else if (nsub==1 && c=='f') {
        var = _FC; // fit curve
    } else if (nsub==1 && c=='k') {
        var = _K;  // internal file index
    } else if (nsub==1 && c=='j') {
        var = _J;  // index of current scan
    } else if (nsub==1 && c=='i') {
        var = _I;  // index of current point
    } else {
        var = _NOREF;
    }
    // newly constructed CRef takes over responsability to destroy CTree's
    tk = _tk;
    tj = _tj;
    ti = _ti;
    if (isindex() && indexed())
        throw string( "invalid cross reference: index cannot be indexed" );
}

void CRef::ref_val( CTOut *ret, const CContext *ctx ) const
{
   // Plain index reference ?
    if ( isindex() ) {
        if     ( var == _K )
            ret->set_u( ctx->k );
        else if ( var == _J )
            ret->set_u( ctx->j );
        else if ( var == _I ) {
            if        ( ctx->dim_1() ) {
                if ( ctx->i==-1 )
                    throw string ( "index i not defined in present context" );
                ret->set_u( ctx->i );
            } else if ( ctx->dim_vi() ) {
                ret->preset_vd( ctx->nv );
                for ( uint i=0; i<ctx->nv; ++i )
                    ret->vd[i] = i;
            } else
                throw string( "BUG ref_val(i) unforeseen reqDim" );
        } else
            throw string( "BUG ref_val(index)" );
        return;
    }

    // get file index k and set file reference f:
    uint k;
    if (tk)
        tk->tree_uival( &k, ctx );
    else
        k = ctx->k;
    if ( k<0 || k>=NOlm::MOM.size())
        throw( "ref_val: invalid file index k=" + strg(k) );
    const CEle* e = &(NOlm::MOM[k]);
    const COlo* f = e->P();

    // get scan index j:
    uint j;
    if ( scanwise() ) {
        if ( tj ) {
            tj->tree_uival( &j, ctx );
        } else {
            j = ctx->j;
            if ( j==(uint)-1 ) {
                if ( e->nScan()!=1 )
                    throw string( "ref_val: in scanwise operation missing j" );
                j = 0;
            }
        }
        if ( j<0 || j>=e->nScan() )
            throw string( "ref_val: invalid scan index j=" + strg(j) );
    } else if ( tj )
        throw string( "ref_val: j=" + ref_info() + 
                      " provided though operarion is not scanwise" );

    // check i:
    if ( !pointwise() && ti )
        throw string( "ref_val: i=" + ref_info() + 
                      " provided though operarion is not pointwise" );

    // Reference for any file type ?
    if ( var==CRef::_NJ ) {
        ret->set_u( e->nScan() );
        return;
    }

    // Reference for data or curve file ?
    const COld* fd;
    const COlc* fc;
    if ( (fd = dynamic_cast<const COld*>(f)) ) {
        if        ( var == _X || var == _Y ) {
            // cout << "CTX " << ctx->info() << "\n";
            if        ( ctx->dim_1() ) {
                // get point index i:
                uint i;
                if ( ti ) {
                    ti->tree_uival( &i, ctx );
                } else {
                    i = ctx->i;
                    if ( i==(uint)-1 )
                        throw string(
                            "ref_val: in pointwise operation missing i" );
                }
                if( i>= fd->VS[j].size() )
                    throw string( "i out of range" );
                if       ( var == _X )
                    ret->set_d( fd->VS[j].x[i] );
                else if ( var == _Y )
                    ret->set_d( fd->VS[j].y[i] );
            } else if ( ctx->dim_vi() ) {
                ret->preset_vd( ctx->nv );
                uint i;
                if ( ti ) {
                    CContext myctx = *ctx;
                    myctx.reqDim = CContext::_1;
                    for( uint ii=0; ii<ctx->nv; ++ii ){
                        myctx.i = ii;
                        ti->tree_uival( &i, &myctx );
                        if( i >= fd->VS[j].size() )
                            throw string( "i out of range" );
                        if      ( var == _X )
                            ret->vd[ii] = fd->VS[j].x[i];
                        else if ( var == _Y )
                            ret->vd[ii] = fd->VS[j].y[i];
                    }
                } else {
                    if ( ctx->nv != fd->VS[j].size() )
                        throw "ref_val " + ref_info() + ": requested vec(" +
                            strg(ctx->nv) + "), found vec(" +
                            strg( fd->VS[j].size() );
                    if      ( var == _X )
                        ret->vd = fd->VS[j].x;
                    else if ( var == _Y )
                        ret->vd = fd->VS[j].y;
                }
            }
        } else if ( var == _Z ) {
            if ( num>= fd->nZ() )
                throw( "invalid z ref(" + ref_info() + ") in data file" );
            ret->set_d( fd->VS[j].z[num] );
        } else if ( var == _NI ) {
            ret->set_d( fd->VS[j].size() );
        } else
            throw( "invalid ref(" + ref_info() + ") in data file" );
    } else if ( (fc = dynamic_cast<const COlc*>(f)) ) {
        if       ( var == _FP ) {
            if ( num<0 || num>= fc->nPar() )
                throw( "invalid p ref(" + ref_info() + ") in curve file" );
            ret->set_d( fc->VC[j].P[num] );
        } else if ( var == _FM ) {
            if ( num<0 || num>= N_FIT_METRICS )
                throw( "invalid fm ref(" + ref_info() + ") in curve file" );
            ret->set_d( fc->VC[j].Quality[num] );
        } else if ( var == _Z ) {
            if ( num>= fc->nZ() )
                throw( "invalid z ref(" + ref_info() + ") in curve file" );
            ret->set_d( fc->VC[j].z[num] );
        } else if ( var == _FC ) {
            ret->set_cr( fc, k, j );
        } else 
            throw( "invalid ref(" + ref_info() + ") in curve file" );
    } else
        throw string( "BUG ref->eval unexpected else" );
}

void CRef::coord( CCoord *ret, uint k_in ) const
{
    if (isindex()) {
        if     ( var == _K )
            *ret = CCoord("k","");
        else if ( var == _J )
            *ret = CCoord("j","");
        else if ( var == _I )
            *ret = CCoord("i","");
        else
            throw string( "BUG coord(index)" );
        return;
    }

    uint k;
    if (tk) {
        // get k from reference :
        CContext ctx( k_in );
        tk->tree_uival(&k, &ctx);
        if (k>=NOlm::MOM.size())
            throw( "coord: file index " + strg(k) + "invalid, maximum is " +
                   strg(NOlm::MOM.size()-1) );
    } else {
        k = k_in;
    }

    const COlo* f = NOlm::MOM[k].P();
    if (!f)
        throw string( "coord: ref where not refering to a file" );

    const COld* fd;
    const COlc* fc;
    if ( (fd = dynamic_cast<const COld*>(f)) ) {
        if (!indatafile())
            throw( "coord: ref " + ref_info() +
                   " not allowed in data file " + strg(k) );
        if       ( var == _X )
            *ret = fd->xco;
        else if  ( var == _Y )
            *ret = fd->yco;
        else if  ( var == _Z ) {
            if (num>=fd->ZCo.size())
                throw( "coord: " + strg(num) + " undefined" );
            *ret = fd->ZCo[num];
        } else if ( var == _NJ )
            *ret = CCoord("#scans", "");
        else if  ( var == _NI )
            *ret = CCoord("#points", "");
        else
            throw( "coord: data:ref(" + ref_info() + ") nyi" );
    } else if ( (fc = dynamic_cast<const COlc*>(f)) ) {
        if (!incurvefile())
            throw( "coord: ref " + ref_info() +
                   " not allowed in curve file " + strg(k) );
        if     ( var == _FP )
            *ret = fc->PCo[num];
        else if ( var == _FM )
            *ret = CCoord("fqi_"+strg(num), ""); // fit quality indicator
        else if ( var == _Z )
            *ret = fc->ZCo[num];
        else if ( var == _FC )
            *ret = CCoord(fc->expr,"");
        else if ( var == _NJ )
            *ret = CCoord("#scans", "");
        else
            throw( "coord: curv:ref " + ref_info() + " nyi" );
    } else
        throw string( "BUG ref->coord unexpected else" );
}

string CRef::ref_info(void) const
{
    string s;
    switch (var) {
    case _X:
        s = "x";  break;
    case _Y:
        s = "y";  break;
    case _FC:
        s = "f";  break;
    case _Z:
        s = "z";  break;
    case _FP:
        s = "p";  break;
    case _FM:
        s = "fm"; break;
    case _K:
        s = "k";  break;
    case _J:
        s = "j";  break;
    case _I:
        s = "i";  break;
    case _NJ:
        s = "nj"; break;
    case _NI:
        s = "ni"; break;
    default:
        s = "?";  break;
    }
    if (numbered()) s += strg(num);
    return s;
}


//***************************************************************************//
//* class CTree                                                             *//
//***************************************************************************//

typedef struct {
    const CScan *sv;
    uint nv;
    double t;
    CContext cvctx;
    CTOut r;
    PTree arg;
} ConvDatTyp;


//! Call-back for integration.

double tree_conv_eval( double tt, void *data )
{
    ConvDatTyp *mydata = (ConvDatTyp*) data;
    if( tt < mydata->sv->x[0] || tt > mydata->sv->x[mydata->nv-1] )
        throw string( "unexpected behavior of gsl: integration variable oor" );
    vector<double> t1(1);
    t1[0] = mydata->t - tt;
    mydata->cvctx.fargs = &t1;
    mydata->arg->tree_val( &(mydata->r), &(mydata->cvctx) );
    return mydata->r.to_d(0) * mydata->sv->intpol( tt );
}


//! Evaluate tree and convert result to integer, for use as index.

void CTree::tree_uival( uint *ret, const CContext *ctx ) const
{
    CTOut val;
    tree_val( &val, ctx );
    // since we have no integer data type, convert real by rounding
    *ret = (uint) ( val.d + 0.5);
}


//! Evaluate tree. Private. User calls pass through API implemented below.

void CTree::tree_val( CTOut *ret, const CContext *ctx ) const
{
    if       ( typ == _OP ) {
        CTOut r[maxarg];
        for ( uint iarg=0; iarg<narg; ++iarg )
            arg[iarg]->tree_val( r+iarg, ctx );
        if       ( narg==1 ) {
            if     ( r[0].is_scalar() )
                ret->set_d( (*(fun->f1))( r[0].d ) );
            else if ( r[0].is_vector() ) {
                int n = r[0].vd.size();
                ret->preset_vd( n );
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = (*(fun->f1))( r[0].vd[i] );
            } else
                throw string( "OP(1) invalid typ" );
        } else if ( narg==2 ) {
            if       ( r[0].is_scalar() && r[1].is_scalar() )
                ret->set_d( (*(fun->f2))( r[0].d, r[1].d ) );
            else if  ( r[0].is_scalar() && r[1].is_vector() ) {
                int n = r[1].vd.size();
                ret->preset_vd( n );
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = (*(fun->f2))( r[0].d, r[1].vd[i] );
            } else if ( r[0].is_vector() && r[1].is_scalar() ) {
                int n = r[0].vd.size();
                ret->preset_vd( n );
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = (*(fun->f2))( r[0].vd[i], r[1].d );
            } else if ( r[0].is_vector() && r[1].is_vector() ) {
                int n = r[0].vd.size();
                if ( r[1].vd.size() != n )
                    throw string( "OP(2) vectors have different size" );
                ret->preset_vd( n );
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = (*(fun->f2))( r[0].vd[i], r[1].vd[i] );
            } else
                throw string( "OP(2) invalid typ" );
        } else if ( narg==3 ) {
            if ( r[0].is_scalar() && r[1].is_scalar() && r[2].is_scalar() ) 
                ret->set_d( (*(fun->f3))( r[0].d, r[1].d, r[2].d ) );
            else {
                int n = -1;
                for ( int iarg=0; iarg<narg; ++iarg ) {
                    if ( r[iarg].is_vector() ) {
                        if ( n==-1 )
                            n = r[iarg].vd.size();
                        else if ( r[iarg].vd.size() != n )
                            throw string( "OP(3) vectors have different size" );
                    }
                }
                if ( n==-1 )
                    throw string( "OP(3) invalid typ" );
                ret->preset_vd( n );
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = (*(fun->f3))(
                        r[0].to_d(i), r[1].to_d(i), r[2].to_d(i) );
            }
        }
    } else if ( typ == _CONV || typ == _DIRAC ) {
        // Return values are requested for n values of _FARG ( = t ):
        if ( !(ctx->fargs) )
            throw string( "BUG: treeval(conv,dirac) with *fargs=0" );
        uint n = ctx->fargs->size();
        ret->preset_vd( n );
        // Resolution is given by context 'ctx':
        uint kconv = setConvK( ctx->k );
        uint jconv = kconv == -1 ? -1 : setConvJ( ctx->k, ctx->j, kconv );
        // cout << "DEBUG k=" << ctx->k << ", j=" << ctx->j << " conv k=" << kconv << ", j=" << jconv << "\n";
        double conv_norm; // normalization constant (integral of resolution)
        double conv_step; // step in x/t, if equidistant
        COld *fv;
        CScan *sv;
        int nv; /* declaring this as unsigned lead to wrong wrapping */
        if( kconv!=-1 ){
            // Check whether return grid is equidistant
            if( n>1 ){
                conv_step = ((*(ctx->fargs))[n-1]-(*(ctx->fargs))[0]) / (n-1);
                if( conv_step<0 )
                    conv_step = 0;
                else {
                    for( int i=0; i<n-1; ++i ) {
                        if( fabs((*(ctx->fargs))[i+1]-(*(ctx->fargs))[i]
                                 -conv_step)>1e-8*conv_step ){
                            conv_step = 0; // not equidistant
                            break;
                        }
                    }
                }
            } else 
                conv_step = 0;
            // Preset fv,sv,nv.
            fv = NOlm::MOM[kconv].D();
            if ( !fv )
                throw "Resolution " + strg(kconv) + "is not a data file";
            sv = &(fv->VS[jconv]);
            nv = sv->size();
            if ( nv < 2 )
                throw "Resolution " + strg(kconv) + ", scan " + strg(jconv) +
                    ", has only " + strg(nv) + " points";
            // Calculate normalization integral.
            // At this occasion, assert that x is ordered,
            conv_norm = 0;
            double dx;
            // special treatment for first and last step:
            dx = sv->x[1] - sv->x[0];
            if ( dx <= 0 )
                throw "Resolution " + strg(kconv) + ", scan " + strg(jconv) +
                    ", not ordered in x";
            conv_norm += sv->y[0] * dx;
            dx = sv->x[nv-1] - sv->x[nv-2];
            if ( dx <= 0 )
                throw "Resolution " + strg(kconv) + ", scan " + strg(jconv) +
                    ", not ordered in x";
            conv_norm += sv->y[nv-1] * dx;
            // the other steps:
            for( int i=1; i<nv-1; ++i ) {
                dx = ( sv->x[i+1] - sv->x[i-1] ) / 2;
                if ( dx <= 0 )
                    throw "Resolution " + strg(kconv) + ", scan " + 
                        strg(jconv) + ", not ordered in x";
                conv_norm += sv->y[i] * dx;
            }
            // check result:
            if ( conv_norm <= 0 )
                throw "Resolution " + strg(kconv) + ", scan " + 
                    strg(jconv) + ", has non-positive norm " + strg(conv_norm);
            // Check whether x is equidistant
            if( conv_step ){
                for( int i=0; i<nv-1; ++i ) {
                    if( fabs(sv->x[i+1]-sv->x[i]-conv_step)>1e-8*conv_step ){
                        conv_step = 0; // not equidistant
                        break;
                    }
                }
            }
            if( !conv_step && !allow_slow_conv )
                throw string(
                    "not equidistant and slow convolution not allowed" );
        }
        // The actual computation is rather different for _CONV and _DIRAC:
        if        ( typ == _CONV ) {
            if( narg!=2 )
                throw string( "BUG: convolution requires 2 args" );
            CTOut r;
            arg[1]->tree_val( &r, ctx );
            if ( !r.is_scalar() )
                throw string( "2nd arg of conv must be scalar shift" );
            double shift = r.d;
            if ( kconv==-1 ) {
                // Copy convolutand (= unconvoluted theory).
                CTOut r;
                // unconvoluted theory given by arg[0]:
                arg[0]->tree_val( &r, ctx );
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = r.to_d(i);
            } else if ( NRead::iofmac("\\cv_mode")==-1 ) {
                CTOut r;
                CContext cvctx(*ctx);
                vector<double> tt;
                cvctx.fargs = &tt;
                // Convolute theory with resolution.
                if( conv_step ){
                    // Accelerated version with equidistant grids 
                    int ntt = n + nv - 1;
                    tt.resize(ntt);
                    for ( int ii=0; ii<ntt; ++ii ) // ii=(nv-1)+i-iv
                        tt[ii] = (*(ctx->fargs))[0] - sv->x[0] - shift +
                            (ii-(nv-1))*conv_step;
                    arg[0]->tree_val( &r, &cvctx );
                    for ( int i=0; i<n; ++i ){
                        ret->vd[i] = 0;
                        for ( int iv=0; iv<nv; ++iv )
                            ret->vd[i] += r.to_d(nv-1+i-iv) * sv->y[iv];
                        ret->vd[i] *= conv_step / conv_norm;
                    }
                } else {
                    // Simplest version: non-equidistant x_out and x_res.
                    for ( int i=0; i<n; ++i )
                        ret->vd[i] = 0;
                    tt.resize(n);
                    for ( int iv=0; iv<nv; ++iv ) {
                        double dx;
                        if     ( iv==0 )
                            dx = sv->x[1] - sv->x[0];
                        else if( iv==nv-1 )
                            dx = sv->x[nv-1] - sv->x[nv-2];
                        else
                            dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
                        for ( int i=0; i<n; ++i )
                            tt[i] = (*(ctx->fargs))[i] - sv->x[iv] - shift;
                        arg[0]->tree_val( &r, &cvctx );
                        for ( int i=0; i<n; ++i )
                            ret->vd[i] += r.to_d(i) * sv->y[iv] *
                                dx / conv_norm;
                    }
                }
            } else {
                // Convolute theory with resolution.
                // Generic integration.
                ConvDatTyp conv_data;
                conv_data.sv = sv;
                conv_data.nv = nv;
                conv_data.arg = arg[0];
                conv_data.cvctx = *ctx;
                for ( int i=0; i<n; ++i ) {
                    conv_data.t = (*(ctx->fargs))[i] - shift;
                    ret->vd[i] =
                        mystd::Integrate( &tree_conv_eval,
                                          (void *) &conv_data,
                                          sv->x[0], sv->x[nv-1],
                                          NRead::iofmac( "\\cv_mode" ),
                                          NRead::dofmac( "\\cv_epsabs" ),
                                          NRead::dofmac( "\\cv_epsrel" ) ) /
                        conv_norm;
                }
            }

        } else if ( typ == _DIRAC ) {
            CTOut r;
            // center of delta function given by arg[0]:
            arg[0]->tree_val( &r, ctx );
            if ( !r.is_scalar() )
                throw string( "dirac function requires scalar argument" );
            if ( kconv==-1 ) {
                // Naive representation of delta function.
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = fabs((*(ctx->fargs))[i]-r.to_d(i))<1e-2 ?
                        1e2 : 0;
            } else {
                // Delta function copies resolution.
                vector<double> tt(n), yy(n);
                // shift in t
                for ( int i=0; i<n; ++i )
                    tt[i] = (*(ctx->fargs))[i]-r.d;
                // interpolation
                sv->intpol( tt, yy );
                // normalization
                for ( int i=0; i<n; ++i )
                    ret->vd[i] = yy[i] / conv_norm;
            }
        } else
            throw string( "FATAL: unexpected typ" );
    } else if ( typ == _VAL ) {
        ret->set_d( val );
    } else if ( typ == _FARG ) {
        if ( !(ctx->fargs) )
            throw string( "BUG: tree_val with *fargs=0" );
        ret->set_vd( *(ctx->fargs) );
    } else if ( typ == _REF ) {
        ref->ref_val( ret, ctx );
    } else if ( typ == _FRF ) {
        // evaluate function reference:
        CTOut frf;
        ref->ref_val( &frf, ctx );
        if( !frf.is_curveref() )
            throw string( "unexpected return type in function reference" );
        // evaluate function argument:
        CTOut rarg;
        arg[0]->tree_val( &rarg, ctx );
        vector<double> vt;
        if ( rarg.is_scalar() ) {
            vt.resize(1);
            vt[0] = rarg.d;
        } else {
            vt.resize(rarg.vd.size());
            vt = rarg.vd;
        }
        // evaluate function:
        CContext cu_ctx( frf.cr.k, frf.cr.j );
        cu_ctx.set_fargs( &vt );
        frf.cr.fc->T->tree_val( ret, &cu_ctx );
        if ( rarg.is_scalar() )
            ret->set_d( ret->vd[0] );
    } else if ( typ == _NOTREE )
        throw string( "BUG: tree_val(tree_typ=NOTREE)" );
    else
        throw string( "BUG: tree_val(tree_typ=UNDEF)" );
}


//! Evaluate tree. Explicit API.

void CTree::tree_point_val( double *ret, uint k, uint j, uint i ) const
{
    //cout << "tree_point_val\n";
    CContext ctx( k, j, i );
    CTOut val;
    tree_val( &val, &ctx );
    if ( !val.is_scalar() )
        throw "tree_point_val got " + val.info();
    *ret = val.d;
}


//! Evaluate tree. Explicit API.

void CTree::tree_vec_val( vector<double> *ret, uint k, uint j ) const
{
    //cout << "tree_vec_val\n";
    CTOut val;
    CContext ctx( k, j );
    ctx.set_nv( ret->size() );
    tree_val( &val, &ctx );
    if ( val.is_vector() )
        (*ret) = val.vd;
    else
        for( uint iv=0; iv<ret->size(); ++iv )
            (*ret)[iv] = val.d;
}


//! Evaluate a curve file's tree, substituting _T and _FP.

void CTree::tree_curve_val_vec(
    vector<double> *ret,const vector<double>& fargs, uint k, uint j
    ) const
{
    //cout << "tree_curve_val\n";
    uint n = fargs.size();
    ret->resize( n );
    CTOut val;
    if        ( !has_farg ) {  // curve does not depend on t
        CContext ctx( k, j );
        tree_val( &val, &ctx );
        if ( !val.is_scalar() )
            throw "tree_curve_val (sca) got " + val.info();
        (*ret)[0] = val.d;
    } else {
        CContext ctx( k, j );
        ctx.set_fargs( &fargs );
        tree_val( &val, &ctx );
        if ( !val.is_vector() )
            throw "tree_curve_val (vec) got " + val.info();
        *ret = val.vd;
    }
}

//! Dito, scalar, not optimized

double CTree::tree_curve_val_sca( const double fargs, uint k, uint j ) const
{
    vector<double> in(1), out(1);
    in[0] = fargs;
    tree_curve_val_vec( &out, in, k, j );
    return out[0];
}

//! Retrieve kconv.

uint CTree::setConvK( uint k )
{
    CEle *e;
    e = &(NOlm::MOM[k]);
    if ( COlc* fc = e->C() )
        return fc->kconv;
    else
        return -1;
}


//! Determine jconv.

uint CTree::setConvJ( uint k, uint j, uint kconv )
{
    CEle *econv, *e;
    if( kconv==-1 ){
        return -1;
    } else {
        // check k,j
        if ( k >= NOlm::MOM.size() )
            throw string( "invalid k" );
        e = &(NOlm::MOM[k]);
        if ( j>=e->nScan() )
            throw string( "invalid j" );
        // check kconv
        if ( kconv >= NOlm::MOM.size() )
            throw string( "invalid kconv" );
        econv = &(NOlm::MOM[kconv]);
        // set jconv
        if      ( econv->nScan()==1 )
            return 0;
        else if ( econv->nScan()==e->nScan() )
            return j;
        else
            throw "cannot convolute file " + strg(k) + " with resolution " +
                strg(kconv) + ": number of scans does not match";
    }
}


//! Compose coordinate names and units according to the tree.

void CTree::coord( CCoord *ret, uint k ) const
{
    CCoord r[maxarg];
    if     ( typ == _OP ) {
        for ( uint iarg=0; iarg<narg; ++iarg )
            arg[iarg]->coord( &(r[iarg]), k );
        if     ( narg==1 )
            *ret = fun->coord( r+0);
        else if ( narg==2 ) 
            *ret = fun->coord( r+0, r+1 );
        else if ( narg==3 ) 
            *ret = fun->coord( r+0, r+1, r+2 );
    } 
    else if ( typ == _CONV )
        arg[0]->coord( ret, k );
    else if ( typ == _DIRAC )
        *ret = CCoord("dirac", "");        
    else if ( typ == _VAL )
        *ret = CCoord(strg(val), "");
    else if  ( typ == _REF )
        ref->coord(ret,k);
    else if  ( typ == _FRF ) {
        // ref->coord(ret,k);
        *ret = CCoord( "f", "" ); // primitiver gehts nicht
    } else 
        throw string( "BUG: tree::coord(tree_typ=UNDEF)" );
}

string CTree::tree_info() const
{
    string ret;
    if       ( typ == _OP ) {
        ret = "op[";
        for ( uint iarg=0; iarg<narg; ++iarg )
            ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : "]" );
        return ret;
    }
    else if ( typ == _CONV )
        return "conv(" + arg[0]->tree_info() + ")";
    else if ( typ == _DIRAC )
        return "dirac(" + arg[0]->tree_info() + ")";
    else if ( typ == _VAL )
        return "val[" + strg(val) + "]";
    else if ( typ == _REF )
        return "ref[" + ref->ref_info() + "]";
    else if ( typ == _FRF )
        return "frf[" + ref->ref_info() + "]";
    else if ( typ == _FARG )
        return "farg";
    else
        return "notree? (BUG ?)";
}


//! How many curve parameters are contained in the tree?

uint CTree::npar() const
{
    uint np=0;
    npar_exec( &np );
    return np;
}

void CTree::npar_exec( uint *np ) const
{
    if      ( typ == _OP || typ == _CONV || typ == _DIRAC )
        for ( uint iarg=0; iarg<narg; ++iarg )
            arg[iarg]->npar_exec( np );
    else if ( typ == _VAL || typ == _FARG || typ == _FRF )
        ; // do nothing ( leave preset value 0 ? )
    else if ( typ == _REF ) {
        if ( ref->var==CRef::_FP ) {
            if (ref->numbered())
                *np = max (*np, ref->num+1);
            else
                *np = 1;
        }
    } else
        throw string( "BUG: npar(UNDEF)" );
}


//***************************************************************************//
//* direct access from main program                                         *//
//***************************************************************************//

void NCalc::Calculator() 
{
    string s;
    PTree T;
    double v;

    if ( NRead::stack_empty() ) {
        // interactive calculator
        while (1) {
            s = sask("calc> ");
            if (s[0]=='q')
                return;
            if (s[0]=='h' || s[0]=='?') {
                printf( "enter valid arithmetic expression, or q to quit\n");
                continue;
            }
            user_xaxparse( s.c_str(), &T );
            T->tree_point_val( &v );
            printf("%.12g\n", v);
        }
    } else {
        // evaluate one expression, with respect to given file selection
        s = NRead::readln("");
        user_xaxparse( s.c_str(), &T );
        NOlm::IterateEle iter;
        CEle *ein;
        if( iter.size()==0 ) {
            T->tree_point_val( &v, -1 );
            printf("%.12g\n", v);
        } else {
            while ( (ein=iter()) ) {
                uint k = iter.SelNo();
                T->tree_point_val( &v, k );
                if ( iter.size()>1 )
                    printf( "f%d: ", k );
                printf("%.12g\n", v);
            }
        }
    }
}
